import typing
from datetime import datetime
from pathlib import Path

import attr
import joblib
import numpy
import pandas

from ..conditions import Conditions
from ..mixtures import Composition, Mixture, Mixtures
from ..optimizer import PervaporationFunction
from ..permeance import Permeance, Units
from ..plotting import plot_graph

PROCESS_MODEL_COLUMNS = [
    "membrane_name",
    "mixture",
    "time",
    "feed_mass",
    "feed_temperature",
    "permeate_temperature",
    "permeate_pressure",
    "composition",
    "composition_type",
    "permeate_composition",
    "permeate_composition_type",
    "partial_flux_1",
    "partial_flux_2",
    "permeance_1",
    "permeance_2",
    "units",
    "feed_evaporation_heat",
    "permeate_condensation_heat",
    "comment",
]


@attr.s(auto_attribs=True)
class ProcessModel:
    """
    Clas for description, working with and storage of process models
    """

    mixture: Mixture
    membrane_name: str
    feed_temperature: typing.List[float]
    feed_compositions: typing.List[Composition]
    permeate_composition: typing.List[Composition]
    permeate_temperature: typing.List[float]
    permeate_pressure: typing.List[float]
    feed_mass: typing.List[float]
    partial_fluxes: typing.List[typing.Tuple[float, float]]
    permeances: typing.List[typing.Tuple[Permeance, Permeance]]
    time: typing.List[float]
    feed_evaporation_heat: typing.List[float]
    permeate_condensation_heat: typing.List[typing.Optional[float]]
    initial_conditions: typing.Optional[Conditions] = None
    permeance_fits: typing.Optional[
        typing.Tuple[PervaporationFunction, PervaporationFunction]
    ] = None
    comments: typing.Optional[str] = None
    membrane_path: typing.Optional[Path] = None

    @staticmethod
    def _generate_process_path(membrane_path: typing.Union[str, Path]) -> Path:
        """
        :param membrane_path: path to an associated membrane
        :return: generates a path to save a process model
        """
        if type(membrane_path) is not Path:
            membrane_path = Path(membrane_path)

        results_path = membrane_path / "results"
        results_path.mkdir(parents=True, exist_ok=True)

        process_path = results_path / ("process_" + str(hash(datetime.now()))[0:4])
        process_path.mkdir(parents=True, exist_ok=False)

        return process_path

    @property
    def get_separation_factor(self) -> typing.List[float]:
        """
        :return: List of separation Factors
        """
        feed = self.feed_compositions
        permeate = self.permeate_composition
        return [
            (permeate[i].first / permeate[i].second) / (feed[i].first / feed[i].second)
            for i in range(len(feed))
        ]

    @property
    def get_psi(self) -> typing.List[float]:
        """
        :return: List of Pervaporation Separation Index (PSI) values
        """
        separation_factor = self.get_separation_factor
        total_flux = [
            sum(self.partial_fluxes[i]) for i in range(len(self.partial_fluxes))
        ]
        return list(numpy.multiply(total_flux, numpy.subtract(separation_factor, 1)))

    @property
    def get_selectivity(self) -> typing.List[float]:
        """
        :return: List of Calculated selectivities
        """
        permeance = self.permeances
        return [
            permeance[i][0].value / permeance[i][1].value for i in range(len(permeance))
        ]

    @classmethod
    def load(cls, process_path: typing.Union[str, Path]) -> "ProcessModel":
        """
        Creates a ProcessModel object from the data in a specified directory
        :param process_path: path to the process folder of a unified format
        :return: ProcessModel object
        """
        if type(process_path) is not Path:
            process_path = Path(process_path)

        process_frame = pandas.read_csv(process_path / "process_model.csv")

        pv_0_filenames = list(
            filter(
                lambda x: x.stem.startswith("pervaporation_function_0"),
                process_path.iterdir(),
            )
        )
        pv_1_filenames = list(
            filter(
                lambda x: x.stem.startswith("pervaporation_function_1"),
                process_path.iterdir(),
            )
        )
        if len(pv_0_filenames) != 1:
            raise FileExistsError(
                "Distinct first pervaporation function file is not found in %s"
                % process_path
            )
        if len(pv_1_filenames) != 1:
            raise FileExistsError(
                "Distinct second pervaporation function file is not found in %s"
                % process_path
            )

        pv_0 = PervaporationFunction.load(pv_0_filenames[0])
        pv_1 = PervaporationFunction.load(pv_1_filenames[0])

        if (process_path / "initial_conditions.ic").exists():
            initial_conditions = joblib.load(process_path / "initial_conditions.ic")
        else:
            initial_conditions = None

        mixture = getattr(Mixtures, process_frame["mixture"].iloc[0])

        if pandas.isna(process_frame["permeate_temperature"].iloc[0]):
            permeate_temperature = None
        else:
            permeate_temperature = process_frame["permeate_temperature"].iloc[0]
        if pandas.isna(process_frame["permeate_pressure"].iloc[0]):
            permeate_pressure = None
        else:
            permeate_pressure = process_frame["permeate_pressure"].iloc[0]

        if (
            process_frame["partial_flux_1"].isna().mean() == 0
            and process_frame["partial_flux_2"].isna().mean() == 0
        ):
            partial_fluxes = []
            for i in range(len(process_frame)):
                partial_fluxes.append(
                    (
                        process_frame["partial_flux_1"].iloc[i],
                        process_frame["partial_flux_2"].iloc[i],
                    )
                )
        else:
            partial_fluxes = None

        if (
            process_frame["permeance_1"].isna().mean() == 0
            and process_frame["permeance_2"].isna().mean() == 0
            and process_frame["units"].isna().mean() == 0
        ):
            permeances = []
            for i in range(len(process_frame)):
                permeances.append(
                    (
                        Permeance(
                            value=process_frame["permeance_1"].iloc[i],
                            units=process_frame["units"].iloc[0],
                        ).convert(
                            to_units=Units.kg_m2_h_kPa,
                            component=mixture.first_component,
                        ),
                        Permeance(
                            value=process_frame["permeance_2"].iloc[i],
                            units=process_frame["units"].iloc[0],
                        ).convert(
                            to_units=Units.kg_m2_h_kPa,
                            component=mixture.second_component,
                        ),
                    )
                )
        else:
            permeances = None

        return cls(
            mixture=mixture,
            membrane_name=process_frame["membrane_name"].iloc[0],
            feed_temperature=process_frame["feed_temperature"],
            feed_compositions=[
                Composition(
                    p=process_frame["composition"].iloc[i],
                    type=process_frame["composition_type"].iloc[i],
                ).to_weight(mixture=mixture)
                for i in range(len(process_frame))
            ],
            permeate_composition=[
                Composition(
                    p=process_frame["permeate_composition"].iloc[i],
                    type=process_frame["permeate_composition_type"].iloc[i],
                ).to_weight(mixture=mixture)
                for i in range(len(process_frame))
            ],
            permeate_temperature=permeate_temperature,
            permeate_pressure=permeate_pressure,
            feed_mass=process_frame["feed_mass"],
            partial_fluxes=partial_fluxes,
            permeances=permeances,
            time=process_frame["time"],
            feed_evaporation_heat=process_frame["feed_evaporation_heat"],
            permeate_condensation_heat=process_frame["permeate_condensation_heat"],
            initial_conditions=initial_conditions,
            permeance_fits=(pv_0, pv_1)
            if (pv_1 is not None and pv_0 is not None)
            else None,
            comments=process_frame["comment"],
            membrane_path=None,
        )

    def save(self, membrane_path: typing.Union[str, Path] = membrane_path) -> None:
        """
        Saves a ProcessModel object to a specified directory of a unified format
        :param membrane_path: path to the associated membrane
        :return: saves the object
        """
        if type(membrane_path) is not Path:
            membrane_path = Path(membrane_path)

        results_path = membrane_path / "results"
        results_path.mkdir(parents=True, exist_ok=True)

        process_path = self._generate_process_path(membrane_path)

        process_frame = pandas.DataFrame(
            {
                "feed_temperature": [t for t in self.feed_temperature],
                "time": [t for t in self.time],
                "composition": [c.first for c in self.feed_compositions],
                "composition_type": [c.type for c in self.feed_compositions],
                "permeate_composition": [c.first for c in self.permeate_composition],
                "permeate_composition_type": [
                    c.type for c in self.permeate_composition
                ],
                "permeate_temperature": [p_t for p_t in self.permeate_temperature],
                "permeate_pressure": [p_p for p_p in self.permeate_pressure],
                "feed_mass": [m for m in self.feed_mass],
                "partial_flux_1": [f[0] for f in self.partial_fluxes],
                "partial_flux_2": [f[1] for f in self.partial_fluxes],
                "permeance_1": [p[0].value for p in self.permeances],
                "permeance_2": [p[1].value for p in self.permeances],
                "units": [p[0].units for p in self.permeances],
                "feed_evaporation_heat": [h for h in self.feed_evaporation_heat],
                "permeate_condensation_heat": [
                    c_h for c_h in self.permeate_condensation_heat
                ],
            }
        )
        process_frame["membrane_name"] = self.membrane_name
        process_frame["mixture"] = self.mixture.name
        process_frame["comment"] = self.comments
        process_frame = process_frame[PROCESS_MODEL_COLUMNS]
        process_frame.to_csv(process_path / "process_model.csv", index=False)

        self.permeance_fits[0].save(
            (
                process_path
                / f"pervaporation_function_0_{self.mixture.first_component.name}.pv"
            )
        )
        self.permeance_fits[1].save(
            (
                process_path
                / f"pervaporation_function_1_{self.mixture.second_component.name}.pv"
            )
        )

        joblib.dump(self.initial_conditions, (process_path / "initial_conditions.ic"))

    def plot(self, y: typing.List, y_label: str = "", curve: bool = 1):
        """
        Draws basic plot of a specified parameter versus process time
        :param y: parameter
        :param y_label: Name of the axis
        :param curve: if True - connects points with a line, if False - draws raw points
        :return: plots a graph
        """

        x = self.time
        if type(y[0]) == tuple:
            first = []
            second = []
            if isinstance(y[0][0], Permeance):
                for m in y:
                    first.append(m[0].value)
                    second.append(m[1].value)
            elif type(y[0][0]) == float or type(y[0][0]) == numpy.float64:
                for m in y:
                    first.append(m[0])
                    second.append(m[1])
            else:
                raise ValueError(f"Unexpected data type {type(y[0][0])}")

            scaling_factor = numpy.floor(numpy.log10((max(first)))) - numpy.floor(
                numpy.log10((max(second)))
            )
            second = [c * (10**scaling_factor) for c in second]
            points = {
                f"First Component - {self.mixture.first_component.name}": (
                    x,
                    first,
                    curve,
                ),
                f"Second Component - {self.mixture.second_component.name}, "
                f"multiplied by {int(10**scaling_factor):.0e} ": (
                    x,
                    second,
                    curve,
                ),
            }

        elif type(y[0]) == float or type(y[0]) == numpy.float64:

            points = {f"{y_label}": (x, y, curve)}

        elif isinstance(y[0], Composition):

            points = {
                f"{y[0].type} fraction of {self.mixture.first_component.name}": (
                    x,
                    [c.first for c in y],
                    curve,
                )
            }
        else:
            raise ValueError(f"Unexpected data type {type(y[0])}")

        plot_graph(
            x_label=f"Process time, hours",
            y_label=y_label,
            points=points,
        )



import typing
from copy import copy
from pathlib import Path

import attr
import joblib
import numpy
from scipy import optimize

from ..diffusion_curve import DiffusionCurve, DiffusionCurveSet
from ..plotting import plot_graph, plot_surface


@attr.s(auto_attribs=True)
class Measurement:
    """
    Parameters:
    x (float): concentration
    t (float): temperature in K
    p (float): Permeance value
    """

    x: float
    t: float
    p: float


@attr.s(auto_attribs=True)
class Measurements:
    """
    List of Measurement objects
    """

    data: typing.List[Measurement]

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, item: int) -> Measurement:
        return self.data[item]

    def __add__(self, other):
        return Measurements(data=self.data + other.data)

    @classmethod
    def from_diffusion_curve_first(cls, curve: DiffusionCurve) -> "Measurements":
        """
        :param curve: Diffusion curve
        :return: Measurements for the first component
        """
        return Measurements(
            data=[
                Measurement(
                    x=curve.feed_compositions[i].first,
                    t=curve.feed_temperature,
                    p=curve.permeances[i][0].value,
                )
                for i in range(len(curve))
            ]
        )

    def append(self, measurement: Measurement) -> None:
        self.data.append(measurement)

    @classmethod
    def from_diffusion_curve_second(cls, curve: DiffusionCurve) -> "Measurements":
        """
        :param curve: Diffusion curve
        :return: Measurements for the second component
        """
        return Measurements(
            data=[
                Measurement(
                    x=curve.feed_compositions[i].first,
                    t=curve.feed_temperature,
                    p=curve.permeances[i][1].value,
                )
                for i in range(len(curve))
            ]
        )

    @classmethod
    def from_diffusion_curves_first(cls, curves: DiffusionCurveSet) -> "Measurements":
        """
        :param curves: Diffusion curves
        :return: Measurements for the first component from multiple diffusion curves
        """
        output = Measurements(data=[])
        for curve in curves:
            output += cls.from_diffusion_curve_first(curve)
        return output

    @classmethod
    def from_diffusion_curves_second(cls, curves: DiffusionCurveSet) -> "Measurements":
        """
        :param curves: Diffusion curves
        :return: Measurements for the second component from multiple diffusion curves
        """
        output = Measurements(data=[])
        for curve in curves:
            output += cls.from_diffusion_curve_second(curve)
        return output


@attr.s(auto_attribs=True)
class PervaporationFunction:
    """
    :param n: max power of temperature independent components in equation
    :param m: max power of temperature dependent components in equation
    :param alpha: multiplication coefficient
    :param a: list of coefficients for temperature independent components
    :param b: list of coefficients for temperature dependent components
    """

    n: int
    m: int

    alpha: float
    a: typing.List[float]
    b: typing.List[float]

    @classmethod
    def from_array(
        cls, array: typing.Union[typing.List[float], numpy.ndarray], n: int, m: int
    ) -> "PervaporationFunction":
        """

        :param array: list of coefficients
        :param n: max power of temperature independent components in equation
        :param m: max power of temperature dependent components in equation
        :return: PervaporationFunction
        """
        assert len(array) == 2 + n + m
        return cls(
            n=n,
            m=m,
            alpha=array[0],
            a=array[1 : n + 1],
            b=array[n + 1 :],
        )

    def __call__(self, x: float, t: float) -> float:
        return self.alpha * (
            numpy.exp(
                sum(self.a[i] * numpy.power(x, i + 1) for i in range(len(self.a)))
                - sum(self.b[i] * numpy.power(x, i) for i in range(len(self.b))) / t
            )
        )

    def __mul__(self, constant: float) -> "PervaporationFunction":
        return PervaporationFunction(
            n=self.n,
            m=self.m,
            alpha=self.alpha * constant,
            a=self.a,
            b=self.b,
        )

    @classmethod
    def load(cls, path: typing.Union[str, Path]) -> "PervaporationFunction":
        """
        :param path: path to binary file with PervaporationFunction
        :return: PervaporationFunction
        """
        if type(path) is not Path:
            path = Path(path)
        return joblib.load(path)

    def save(self, path: typing.Union[str, Path]) -> None:
        """
        :param path: path where to save binary file with PervaporationFunction
        """
        joblib.dump(self, path)

    def plot(
        self,
        experimental_data: typing.Optional[Measurements] = None,
        temperature: typing.Optional[float] = None,
        concentration: typing.Tuple[float, float] = None,
    ):
        points = {}
        x, t, p = 0, 0, 0

        if experimental_data is not None:
            x = [m.x for m in experimental_data]
            t = [m.t for m in experimental_data]
            p = [m.p for m in experimental_data]
            x_max = max(x)
            x_min = min(x)
            t_max = max(t)
            t_min = min(t)
            if temperature is not None and not (temperature in set(t)):
                raise ValueError(f"No experimental points at {temperature} K available")
            if temperature is not None:
                x_selected = []
                p_selected = []
                for i in range(len(t)):
                    if t[i] == temperature:
                        x_selected.append(x[i])
                        p_selected.append(p[i])
                x = x_selected
                p = p_selected

            points["Experiment"] = (x, p, False)
            if len(set(t)) == 1:
                temperature = set(t).pop()

        else:
            if concentration is None:
                x_max = 1
                x_min = 0
                t_max = 373.15
                t_min = 293.15
            else:
                if max(concentration) > 1 or min(concentration) < 0:
                    raise ValueError("Concentration must be in [0,1] range")
                x_max = max(concentration)
                x_min = min(concentration)
                t_max = 373.15
                t_min = 293.15

        x_v = numpy.linspace(x_min, x_max, num=50)

        if temperature is not None:
            p_fit = [self(x_v[i], temperature) for i in range(len(x_v))]
            points["Fit"] = (x_v, p_fit, True)
            plot_graph(
                x_label="First component fraction in feed",
                y_label="Permeance",
                points=points,
                title=f"Temperature {temperature} K",
            )
            return

        plot_surface(
            condition=experimental_data is not None,
            function=self,
            x=x,
            t=t,
            p=p,
            t_min=t_min,
            t_max=t_max,
            x_v=x_v,
        )


def _suggest_n_m(
    data: Measurements, n: typing.Optional[int] = None, m: typing.Optional[int] = None
) -> typing.Tuple[int, int]:
    if n is None:
        suggested_n = int(numpy.sqrt(len(data)))
    else:
        suggested_n = n
    if m is None:
        suggested_m = int(numpy.sqrt(len(data)))
    else:
        suggested_m = m
    return suggested_n, suggested_m


def get_initial_guess(n: int, m: int) -> typing.List[float]:
    return [1] * (2 + n + m)


def objective(data: Measurements, params: typing.List[float], n: int, m: int) -> float:
    error = 0
    foo = PervaporationFunction.from_array(array=params, n=n, m=m)
    for d in data:
        error += (foo(d.x, d.t) - d.p) ** 2
    return numpy.sqrt(error / len(data))


def fit(
    data: Measurements,
    n: typing.Optional[int] = None,
    m: typing.Optional[int] = None,
    include_zero: bool = False,
    component_index: int = 0,
) -> PervaporationFunction:
    """
    Fit PervaporationFunction to given data with pre-defined parameters
    :param data: List of Measurements
    :param n: max power of temperature independent components in equation
    :param m: max power of temperature dependent components in equation
    :param include_zero: include zero-datapoint or not
    :param component_index: index of component to fit
    :return: PervaporationFunction
    """
    _n, _m = _suggest_n_m(data, n, m)
    if component_index not in {0, 1}:
        raise ValueError("Index should be either 0 or 1")

    _data = copy(data)

    if include_zero:
        unique_temperatures = set([m.t for m in data])
        for t in unique_temperatures:
            _data.append(
                Measurement(
                    x=component_index,
                    t=t,
                    p=0,
                )
            )

    result = optimize.minimize(
        lambda params: objective(_data, params, n=_n, m=_m),
        x0=numpy.array([0] * (2 + _n + _m)),
        method="Powell",
    )
    return PervaporationFunction.from_array(array=result.x, n=_n, m=_m)


def find_best_fit(
    data: Measurements,
    include_zero: bool = False,
    component_index: int = 0,
    n: typing.Optional[int] = None,
    m: typing.Optional[int] = None,
) -> PervaporationFunction:
    """
    Finds best fit of PervaporationFunction to given data
    :param data: List of Measurements
    :param include_zero: include zero-datapoint or not
    :param component_index: index of component to fit
    :param n: max power of temperature independent components in equation
    :param m: max power of temperature dependent components in equation
    :return: PervaporationFunction
    """
    max_power_n = min(5, round(numpy.power(len(data), 0.5)))
    max_power_m = min(5, max(1, len(set([measurement.t for measurement in data])) - 1))

    if n is None:
        n_tries = list(range(max_power_n))
    else:
        n_tries = list(range(n + 1))
        if n >= len(data):
            print(
                "n is more or equal to the number of available points, you may be over-fitting"
            )
    if m is None:
        m_tries = list(range(max_power_m))
    else:
        m_tries = list(range(m + 1))
        if m >= len(set([measurement.t for measurement in data])):
            print(
                "m is more or equal to the number of available temperature points, you may be over-fitting"
            )

    best_curve = None
    best_loss = numpy.inf

    for n in n_tries:
        for m in m_tries:
            curve = fit(
                data,
                n=n,
                m=m,
                include_zero=include_zero,
                component_index=component_index,
            )
            loss = sum(
                [
                    (curve(measurement.x, measurement.t) - measurement.p) ** 2
                    for measurement in data
                ]
            )

            if loss < best_loss:
                best_curve = curve
                best_loss = loss

    return best_curve



import typing
from pathlib import Path

import attr
import numpy
import pandas

from ..mixtures import (Composition, CompositionType, Mixture, Mixtures,
                        get_nrtl_partial_pressures)
from ..permeance import Permeance, Units
from ..plotting import plot_graph

DC_SET_COLUMNS = [
    "curve_id",
    "membrane_name",
    "mixture",
    "feed_temperature",
    "permeate_temperature",
    "permeate_pressure",
    "composition",
    "composition_type",
    "partial_flux_1",
    "partial_flux_2",
    "permeance_1",
    "permeance_2",
    "units",
    "comment",
]


@attr.s(auto_attribs=True)
class DiffusionCurve:
    """
    The class for Diffusion Curves - dependency of selective-transport properties
    on feed composition, and process temperature
    """

    mixture: Mixture
    membrane_name: str
    feed_temperature: float
    feed_compositions: typing.List[Composition]
    partial_fluxes: typing.Optional[typing.List[typing.Tuple[float, float]]] = None
    permeate_temperature: typing.Optional[float] = None
    permeate_pressure: typing.Optional[float] = None
    permeances: typing.Optional[typing.List[typing.Tuple[Permeance, Permeance]]] = None
    comments: typing.Optional[str] = None

    def __attrs_post_init__(self):

        if self.partial_fluxes is None and self.permeances is not None:
            """
            Calculation of Partial Fluxes values if they are not specified
            Permeance values are converted to kg/(m2*h*kPa),
            The permeate temperature and pressure parameters are ignored, test_components' pressure in permeate
            is considered zero
            :return a list of Partial fluxes for each component tuple(Ji,Jj) at each concentration
            """
            feed_partial_pressures = [
                get_nrtl_partial_pressures(
                    self.feed_temperature, self.mixture, composition
                )
                for composition in self.feed_compositions
            ]

            self.permeances = [
                (
                    self.permeances[i][0].convert(
                        to_units=Units.kg_m2_h_kPa,
                        component=self.mixture.first_component,
                    ),
                    self.permeances[i][1].convert(
                        to_units=Units.kg_m2_h_kPa,
                        component=self.mixture.second_component,
                    ),
                )
                for i in range(len(self.permeances))
            ]

            self.partial_fluxes = [
                (
                    self.permeances[i][0].value * feed_partial_pressures[i][0],
                    self.permeances[i][1].value * feed_partial_pressures[i][1],
                )
                for i in range(len(self.permeances))
            ]

        if self.permeances is None and self.partial_fluxes is not None:
            """
            Calculation of Permeance values if they are not specified
            Permeance values are in kg/(m2*h*kPa)
            If needed, Permeance values may be converted using Permeance.converter()
            :return a list of Permeances for each component tuple(Pi,Pj) at each concentration
            """
            permeate_compositions = self.permeate_composition
            feed_partial_pressures = [
                get_nrtl_partial_pressures(
                    self.feed_temperature, self.mixture, composition
                )
                for composition in self.feed_compositions
            ]
            if self.permeate_temperature is None and self.permeate_pressure is None:
                self.permeances = [
                    (
                        Permeance(
                            value=self.partial_fluxes[i][0]
                            / feed_partial_pressures[i][0]
                        ),
                        Permeance(
                            value=self.partial_fluxes[i][1]
                            / feed_partial_pressures[i][1]
                        ),
                    )
                    for i in range(len(self.feed_compositions))
                ]
            elif (
                self.permeate_temperature is not None and self.permeate_pressure is None
            ):
                permeate_partial_pressures = [
                    get_nrtl_partial_pressures(
                        self.permeate_temperature, self.mixture, composition
                    )
                    for composition in permeate_compositions
                ]
                self.permeances = [
                    (
                        Permeance(
                            value=self.partial_fluxes[i][0]
                            / (
                                feed_partial_pressures[i][0]
                                - permeate_partial_pressures[i][0]
                            )
                        ),
                        Permeance(
                            value=self.partial_fluxes[i][1]
                            / (
                                feed_partial_pressures[i][1]
                                - permeate_partial_pressures[i][1]
                            )
                        ),
                    )
                    for i in range(len(self.feed_compositions))
                ]

            elif (
                self.permeate_pressure is not None and self.permeate_temperature is None
            ):
                permeate_partial_pressures = [
                    (
                        self.permeate_pressure
                        * composition.to_molar(self.mixture).first,
                        self.permeate_pressure
                        * composition.to_molar(self.mixture).second,
                    )
                    for composition in permeate_compositions
                ]
                self.permeances = [
                    (
                        Permeance(
                            value=self.partial_fluxes[i][0]
                            / (
                                feed_partial_pressures[i][0]
                                - permeate_partial_pressures[i][0]
                            )
                        ),
                        Permeance(
                            value=self.partial_fluxes[i][1]
                            / (
                                feed_partial_pressures[i][1]
                                - permeate_partial_pressures[i][1]
                            )
                        ),
                    )
                    for i in range(len(self.feed_compositions))
                ]
            else:
                raise ValueError(
                    "Either permeate temperature or permeate pressure could be stated not both"
                )
        elif self.permeances is None and self.partial_fluxes is None:
            raise ValueError(
                "Either Permeances or Fluxes must be specified as functions of feed composition"
            )
        else:
            self.permeances = [
                (
                    self.permeances[i][0].convert(
                        to_units=Units.kg_m2_h_kPa,
                        component=self.mixture.first_component,
                    ),
                    self.permeances[i][1].convert(
                        to_units=Units.kg_m2_h_kPa,
                        component=self.mixture.second_component,
                    ),
                )
                for i in range(len(self.permeances))
            ]

    def __len__(self):
        return len(self.feed_compositions)

    def plot(self, y: typing.List, y_label: str = "", curve: bool = 1):
        """
        Draws basic plot of a specified parameter versus first component fraction in the feed
        :param y: parameter
        :param y_label: Name of the axis
        :param curve: if True - connects points with a line, if False - draws raw points
        :return: plots a graph
        """

        x = [c.first for c in self.feed_compositions]
        if type(y[0]) == tuple:
            first = []
            second = []
            if isinstance(y[0][0], Permeance):
                for m in y:
                    first.append(m[0].value)
                    second.append(m[1].value)
            elif type(y[0][0]) == float or type(y[0][0]) == numpy.float64:
                for m in y:
                    first.append(m[0])
                    second.append(m[1])
            else:
                raise ValueError(f"Unexpected default_membranes type {type(y[0][0])}")

            scaling_factor = numpy.floor(numpy.log10((max(first)))) - numpy.floor(
                numpy.log10((max(second)))
            )
            second = [c * (10**scaling_factor) for c in second]
            points = {
                f"First Component - {self.mixture.first_component.name}": (
                    x,
                    first,
                    curve,
                ),
                f"Second Component - {self.mixture.second_component.name}, "
                f"multiplied by {int(10**scaling_factor):.0e} ": (
                    x,
                    second,
                    curve,
                ),
            }

        elif type(y[0]) == float or type(y[0]) == numpy.float64:

            points = {f"{y_label}": (x, y, curve)}

        elif isinstance(y[0], Composition):
            points = {
                f"{y[0].type} fraction of {self.mixture.first_component.name}": (
                    x,
                    [c.first for c in y],
                    curve,
                )
            }
        else:
            raise ValueError(f"Unexpected default_membranes type {type(y[0])}")

        plot_graph(
            x_label=f"{self.mixture.first_component.name}, {self.feed_compositions[0].type} %",
            y_label=y_label,
            points=points,
        )

    @property
    def permeate_composition(self) -> typing.List[Composition]:
        """
        Calculation of permeate compositions at each concentration
        :return - List of permeate weight compositions
        """
        return [
            Composition(
                self.partial_fluxes[i][0] / (sum(self.partial_fluxes[i])),
                type=CompositionType.weight,
            )
            for i in range(len(self.feed_compositions))
        ]

    @property
    def get_separation_factor(self) -> typing.List[float]:
        """
        Calculation of separation factor at each concentration
        :return - List of separation factors
        """
        permeate_composition = self.permeate_composition
        feed_composition = self.feed_compositions
        return [
            (permeate_composition[i].first / permeate_composition[i].second)
            / (feed_composition[i].first / feed_composition[i].second)
            for i in range(len(self.feed_compositions))
        ]

    @property
    def get_psi(self) -> typing.List[float]:
        """
        Calculation of Pervaporation Separation Index (PSI) at each concentration
        :return - List of PSI
        """
        total_flux = [
            sum(self.partial_fluxes[i]) for i in range(len(self.feed_compositions))
        ]
        separation_factor = self.get_separation_factor
        return numpy.multiply(total_flux, numpy.subtract(separation_factor, 1))

    @property
    def get_permeances(self) -> typing.List[typing.Tuple[Permeance, Permeance]]:
        """
        Calculates permenace of both components based on the partial flux values, mixture, feed  and permeate parameters
        :return: Permeances of each component at each feed concentration as [(P1,P2),...]
        """
        if self.permeances is not None:
            return self.permeances
        else:
            permeate_compositions = self.permeate_composition
            feed_partial_pressures = [
                get_nrtl_partial_pressures(
                    self.feed_temperature, self.mixture, composition
                )
                for composition in self.feed_compositions
            ]
            if self.permeate_temperature is None:
                self.permeances = [
                    (
                        Permeance(
                            value=self.partial_fluxes[i][0]
                            / feed_partial_pressures[i][0]
                        ),
                        Permeance(
                            value=self.partial_fluxes[i][1]
                            / feed_partial_pressures[i][1]
                        ),
                    )
                    for i in range(len(self.feed_compositions))
                ]
            else:
                permeate_partial_pressures = [
                    get_nrtl_partial_pressures(
                        self.permeate_temperature, self.mixture, composition
                    )
                    for composition in permeate_compositions
                ]
                self.permeances = [
                    (
                        Permeance(
                            value=self.partial_fluxes[i][0]
                            / (
                                feed_partial_pressures[i][0]
                                - permeate_partial_pressures[i][0]
                            )
                        ),
                        Permeance(
                            value=self.partial_fluxes[i][1]
                            / (
                                feed_partial_pressures[i][1]
                                - permeate_partial_pressures[i][1]
                            )
                        ),
                    )
                    for i in range(len(self.feed_compositions))
                ]

    @property
    def get_selectivity(
        self,
    ) -> typing.List[float]:
        """
        Calculation of selectivity at each concentration
        :return - List of selectivities (Permeances are in SI by default)
        """
        permeances = self.permeances
        return [
            permeances[i][0].convert("SI", self.mixture.first_component).value
            / permeances[i][1].convert("SI", self.mixture.second_component).value
            for i in range(len(self.feed_compositions))
        ]

    @classmethod
    def from_frame(cls, data: pandas.DataFrame) -> "DiffusionCurve":
        # TODO: docstring
        mixture = getattr(Mixtures, data["mixture"].iloc[0])

        if (
            data["partial_flux_1"].isna().mean() == 0
            and data["partial_flux_2"].isna().mean() == 0
        ):
            partial_fluxes = []
            for i in range(len(data)):
                partial_fluxes.append(
                    (
                        data["partial_flux_1"].iloc[i],
                        data["partial_flux_2"].iloc[i],
                    )
                )
        else:
            partial_fluxes = None

        if (
            data["permeance_1"].isna().mean() == 0
            and data["permeance_2"].isna().mean() == 0
            and data["units"].isna().mean() == 0
        ):
            permeances = []
            for i in range(len(data)):
                permeances.append(
                    (
                        Permeance(
                            value=data["permeance_1"].iloc[i],
                            units=data["units"].iloc[0],
                        ).convert(
                            to_units=Units.kg_m2_h_kPa,
                            component=mixture.first_component,
                        ),
                        Permeance(
                            value=data["permeance_2"].iloc[i],
                            units=data["units"].iloc[0],
                        ).convert(
                            to_units=Units.kg_m2_h_kPa,
                            component=mixture.second_component,
                        ),
                    )
                )
        else:
            permeances = None

        if partial_fluxes is None and permeances is None:
            raise ValueError("Partial fluxes and permeances are not provided")

        if pandas.isna(data["permeate_temperature"].iloc[0]):
            permeate_temperature = None
        else:
            permeate_temperature = data["permeate_temperature"].iloc[0]
        if pandas.isna(data["permeate_pressure"].iloc[0]):
            permeate_pressure = None
        else:
            permeate_pressure = data["permeate_pressure"].iloc[0]

        return DiffusionCurve(
            mixture=mixture,
            membrane_name=data["membrane_name"].iloc[0],
            feed_temperature=data["feed_temperature"].iloc[0],
            feed_compositions=[
                Composition(
                    p=data["composition"].iloc[i], type=data["composition_type"].iloc[i]
                ).to_weight(mixture=mixture)
                for i in range(len(data))
            ],
            partial_fluxes=partial_fluxes,
            permeate_temperature=permeate_temperature,
            permeate_pressure=permeate_pressure,
            permeances=permeances,
            comments=data["comment"].iloc[0],
        )

    def save(self, path: typing.Union[str, Path]) -> None:
        """
        Saves the curve to a specified path in the form of a .csv file
        :param path: path to the directory
        :return: None, the object is saved in a specified directory
        """
        output = pandas.DataFrame(
            {
                "composition": [c.p for c in self.feed_compositions],
                "composition_type": [c.type for c in self.feed_compositions],
                "partial_flux_1": [f[0] for f in self.partial_fluxes],
                "partial_flux_2": [f[1] for f in self.partial_fluxes],
                "permeance_1": [p[0].value for p in self.permeances],
                "permeance_2": [p[1].value for p in self.permeances],
                "units": [p[0].units for p in self.permeances],
            }
        )
        output["curve_id"] = "1"
        output["membrane_name"] = self.membrane_name
        output["mixture"] = self.mixture.name
        output["feed_temperature"] = self.feed_temperature
        output["permeate_temperature"] = self.permeate_temperature
        output["permeate_pressure"] = self.permeate_pressure
        output["comment"] = self.comments
        output = output[DC_SET_COLUMNS]
        output.to_csv(path, index=False)


@attr.s(auto_attribs=True)
class DiffusionCurveSet:
    """
    A class for storing and working with multiple diffusion curves related to the same set of experiments.
    """

    name: str
    diffusion_curves: typing.List[DiffusionCurve]

    def __getitem__(self, item):
        return self.diffusion_curves[item]

    @classmethod
    def load(cls, path: Path) -> "DiffusionCurveSet":
        data = pandas.read_csv(path)
        name = path.stem

        if list(data.columns) != DC_SET_COLUMNS:
            raise ValueError(
                "Incorrect default_membranes: %s at %s" % (list(data.columns), path)
            )

        raw_diffusion_curves = data.groupby("curve_id")
        diffusion_curves = []
        for _, curve in raw_diffusion_curves:
            diffusion_curves.append(DiffusionCurve.from_frame(curve))

        return cls(
            name=name,
            diffusion_curves=diffusion_curves,
        )



from .mixture import (Composition, CompositionType, Mixture,
                      get_nrtl_partial_pressures)
from .mixtures import Mixtures

__all__ = [
    "Mixture",
    "Mixtures",
    "Composition",
    "get_nrtl_partial_pressures",
    "CompositionType",
]



import typing

import attr

R = 8.314462


class VPConstantsType:
    antoine: str = "antoine"
    frost: str = "frost"


@attr.s(auto_attribs=True)
class VaporPressureConstants:
    a: float = attr.ib(converter=lambda value: float(value))  # type: ignore
    b: float = attr.ib(converter=lambda value: float(value))  # type: ignore
    c: float = attr.ib(converter=lambda value: float(value))  # type: ignore
    type: str = attr.ib(
        default=VPConstantsType.antoine,
        converter=lambda x: getattr(VPConstantsType, x)
        if x is not None
        else VPConstantsType.antoine,
    )  # type: ignore


@attr.s(auto_attribs=True)
class NRTLParameters:
    g12: float
    g21: float
    alpha12: float
    alpha21: typing.Optional[float] = None
    a12: typing.Optional[float] = 0
    a21: typing.Optional[float] = 0


@attr.s(auto_attribs=True)
class HeatCapacityConstants:
    a: float = attr.ib(converter=lambda value: float(value))  # type: ignore
    b: float = attr.ib(converter=lambda value: float(value))  # type: ignore
    c: float = attr.ib(converter=lambda value: float(value))  # type: ignore
    d: float = attr.ib(converter=lambda value: float(value))  # type: ignore



from pyvaporation.utils import HeatCapacityConstants, VaporPressureConstants

from .component import Component


class Components:
    """
    A class with pre-defined components with constants validated against literature data
    For details see ./tests/test_components/test_default_components.py
    """

    H2O: Component = Component(
        name="H2O",
        molecular_weight=18.02,
        vapour_pressure_constants=VaporPressureConstants(
            a=7.20389,
            b=-1733.926,
            c=-39.485,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=32.2,
            b=1.924e-3,
            c=1.055e-5,
            d=-3.596e-9,
        ),
    )

    MeOH: Component = Component(
        name="MeOH",
        molecular_weight=32.04,
        vapour_pressure_constants=VaporPressureConstants(
            a=7.2209903,
            b=-1590.15535,
            c=-32.77001,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=89.26344,
            b=-1.66291e-1,
            c=1.84499e-4,
            d=9.44251e-7,
        ),
    )

    EtOH: Component = Component(
        name="EtOH",
        molecular_weight=46.07,
        vapour_pressure_constants=VaporPressureConstants(
            a=7.24677,
            b=-1598.673,
            c=-46.424,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=147.815652,
            b=-0.6732612305,
            c=0.001889017424,
            d=0,
        ),
    )

    iPOH: Component = Component(
        name="iPOH",
        molecular_weight=60.10,
        vapour_pressure_constants=VaporPressureConstants(
            a=6.861,
            b=-1357.427,
            c=-75.814,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=53.8206447,
            b=7.92412193e-1,
            c=-4.56017218e-3,
            d=1.01887619e-5,
        ),
    )

    MTBE: Component = Component(
        name="MTBE",
        molecular_weight=88.15,
        vapour_pressure_constants=VaporPressureConstants(
            a=6.050931522,
            b=-1139.816725,
            c=-46.15171,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=147.329712,
            b=-9.7850807e-2,
            c=9.216480e-4,
            d=-4.75200e-7,
        ),
    )

    ETBE: Component = Component(
        name="ETBE",
        molecular_weight=102.17,
        vapour_pressure_constants=VaporPressureConstants(
            a=6.0703,
            b=-1206.874,
            c=-49.19,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=18.79,
            b=1.251,
            c=-3.015e-3,
            d=3.637e-6,
        ),
    )

    DME: Component = Component(
        name="DME",
        molecular_weight=90.12,
        vapour_pressure_constants=VaporPressureConstants(
            a=5.83775,
            b=-1260.52,
            c=-37.322,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=1.43841e2,
            b=3.88448e-1,
            c=-1.49740e-3,
            d=2.45328e-6,
        ),
    )

    DMC: Component = Component(
        name="DMC",
        molecular_weight=90.08,
        vapour_pressure_constants=VaporPressureConstants(
            a=5.78894,
            b=-1049.375,
            c=-85.977,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=3.29155e2,
            b=-1.54438,
            c=4.42614e-3,
            d=-3.66196e-6,
        ),
    )

    CycloHexane: Component = Component(
        name="CycloHexane",
        molecular_weight=84.16,
        vapour_pressure_constants=VaporPressureConstants(
            a=5.97636182,
            b=-1206.47,
            c=-49.864,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=43.37101,
            b=0.380403,
            c=0,
            d=0,
        ),
    )

    Benzene: Component = Component(
        name="Benzene",
        molecular_weight=78.11,
        vapour_pressure_constants=VaporPressureConstants(
            a=6.00477182,
            b=-1196.76,
            c=-53.839,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=1.18680e2,
            b=-1.01465e-1,
            c=5.41068e-4,
            d=0,
        ),
    )

    Toluene: Component = Component(
        name="Toluene",
        molecular_weight=92.14,
        vapour_pressure_constants=VaporPressureConstants(
            a=6.0854,
            b=-1348.77,
            c=-53.024,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=1.87438e2,
            b=-7.30265e-1,
            c=2.96136e-3,
            d=-2.86617e-6,
        ),
    )

    AceticAcid: Component = Component(
        name="Acetic acid",
        molecular_weight=60.05,
        vapour_pressure_constants=VaporPressureConstants(
            a=6.68206,
            b=-1642.54,
            c=-39.764,
        ),
        heat_capacity_constants=HeatCapacityConstants(
            a=-3.891576e3,
            b=3.640888e1,
            c=-1.099113e-1,
            d=1.1060630e-4,
        ),
    )



import typing
from pathlib import Path

import attr
import pandas

from ..components import Component, Components
from ..permeance import Permeance, Units

IDEAL_EXPERIMENT_COLUMNS = [
    "name",
    "temperature",
    "component",
    "activation_energy",
    "permeance",
    "units",
    "comment",
]


@attr.s(auto_attribs=True)
class IdealExperiment:
    """
    Class for specification of Ideal experiments, where Permeance is assumed to be constant
    over considered composition range
    """

    name: str
    temperature: float
    component: Component
    permeance: Permeance
    activation_energy: typing.Optional[float] = None
    comment: typing.Optional[str] = None

    @classmethod
    def from_dict(
        cls, d: typing.Mapping[str, typing.Union[str, float]]
    ) -> "IdealExperiment":
        """
        Reads IdealExperiment objects from dictionary
        :param d: Mapping of IdealExperiment parameters
        :return: IdealExperiment object with parameters specified in mapping
        """
        component = getattr(Components, d["component"])
        permeance = Permeance(value=d["permeance"], units=d["units"]).convert(
            to_units=Units.kg_m2_h_kPa, component=component
        )
        return cls(
            name=d["name"],
            temperature=d["temperature"],
            component=component,
            permeance=permeance,
            activation_energy=d["activation_energy"],
            comment=d["comment"],
        )


@attr.s(auto_attribs=True)
class IdealExperiments:
    """
    A class for storing multiple IdealExperiments related to a particular membrane
    """

    experiments: typing.List[IdealExperiment]

    def __len__(self):
        return len(self.experiments)

    @classmethod
    def from_csv(cls, path: typing.Union[str, Path]) -> "IdealExperiments":
        """
        Generates IdealExperiments object from the .csv file
        :param path: Path to the file
        :return: IdealExperiments object with IdealExperiments described in a .csv file
        """
        frame = pandas.read_csv(path)
        if list(frame.columns) != IDEAL_EXPERIMENT_COLUMNS:
            raise ValueError("Incorrect columns: %s" % list(frame.columns))

        experiments = []
        for _, row in frame.iterrows():
            experiments.append(IdealExperiment.from_dict(row.to_dict()))

        return IdealExperiments(experiments=experiments)



import datetime
import typing
from datetime import datetime

import attr
import numpy

from ..conditions import Conditions
from ..diffusion_curve import DiffusionCurve, DiffusionCurveSet
from ..membrane import Membrane
from ..mixtures import (Composition, CompositionType, Mixture,
                        get_nrtl_partial_pressures)
from ..optimizer import Measurements, find_best_fit
from ..permeance import Permeance, Units
from ..process import ProcessModel
from ..utils import R


def get_permeate_composition_from_fluxes(
    fluxes: typing.Tuple[float, float],
) -> Composition:
    return Composition(
        p=fluxes[0] / sum(fluxes),
        type=CompositionType.weight,
    )


@attr.s(auto_attribs=True)
class Pervaporation:
    membrane: Membrane
    mixture: Mixture

    def get_partial_fluxes_from_permeate_composition(
        self,
        first_component_permeance: Permeance,
        second_component_permeance: Permeance,
        permeate_composition: Composition,
        feed_composition: Composition,
        feed_temperature: float,
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
    ) -> typing.Tuple[float, float]:
        """
        Calculates partial fluxes at a given Permeate composition, accounting for the driving force change
        Either permeate temperature or permeate pressure could be stated
        :param first_component_permeance - Permeance of the first test_components
        :param second_component_permeance - Permeance of the second test_components
        :param permeate_composition - permeate composition
        :param feed_composition - feed composition
        :param feed_temperature - temperature, K
        :param permeate_temperature - permeate temperature, K , if not specified permeate pressure is considered 0 kPa
        :param permeate_pressure - permeate pressure, kPa , if not specified permeate pressure is considered 0 kPa
        """

        feed_nrtl_partial_pressures = get_nrtl_partial_pressures(
            feed_temperature, self.mixture, feed_composition
        )
        if permeate_temperature is None and permeate_pressure is None:
            permeate_nrtl_partial_pressures = (0, 0)

        elif permeate_temperature is not None and permeate_pressure is None:
            permeate_nrtl_partial_pressures = get_nrtl_partial_pressures(
                permeate_temperature, self.mixture, permeate_composition
            )
        elif permeate_pressure is not None and permeate_temperature is None:
            permeate_nrtl_partial_pressures = (
                permeate_pressure * permeate_composition.first,
                permeate_pressure * permeate_composition.second,
            )

        else:
            raise ValueError(
                "Either permeate temperature or permeate pressure could be stated not both"
            )

        return (
            first_component_permeance.value
            * (feed_nrtl_partial_pressures[0] - permeate_nrtl_partial_pressures[0]),
            second_component_permeance.value
            * (feed_nrtl_partial_pressures[1] - permeate_nrtl_partial_pressures[1]),
        )

    def calculate_partial_fluxes(
        self,
        feed_temperature: float,
        composition: Composition,
        precision: float = 3e-4,
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
        first_component_permeance: typing.Optional[Permeance] = None,
        second_component_permeance: typing.Optional[Permeance] = None,
    ) -> typing.Tuple[float, float]:
        """
        Calculates partial fluxes of the test_components at specified conditions.
        Either permeate temperature or permeate pressure could be stated
        :param feed_temperature: Feed temperature, K
        :param composition: Feed composition
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :param permeate_temperature: Permeate temperature, if not specified permeate pressure is set to 0 kPa
        :param permeate_pressure - permeate pressure, kPa , if not specified permeate pressure is considered 0 kPa
        :param first_component_permeance: Permeance of the first test_components, if not specified is calculated
        :param second_component_permeance: Permeance of the second test_components, if not specified is calculated
        :return: Partial fluxes of test_components as a tuple
        """
        if second_component_permeance is None or first_component_permeance is None:
            first_component_permeance = self.membrane.get_permeance(
                feed_temperature, self.mixture.first_component
            ).convert(
                to_units=Units().kg_m2_h_kPa, component=self.mixture.first_component
            )
            second_component_permeance = self.membrane.get_permeance(
                feed_temperature, self.mixture.second_component
            ).convert(
                to_units=Units().kg_m2_h_kPa, component=self.mixture.second_component
            )

        initial_fluxes: typing.Tuple[float, float] = numpy.multiply(
            (first_component_permeance.value, second_component_permeance.value),
            get_nrtl_partial_pressures(feed_temperature, self.mixture, composition),
        )
        permeate_composition = get_permeate_composition_from_fluxes(initial_fluxes)

        d = 1
        while d >= precision:
            try:
                permeate_composition_new = get_permeate_composition_from_fluxes(
                    self.get_partial_fluxes_from_permeate_composition(
                        first_component_permeance=first_component_permeance,
                        second_component_permeance=second_component_permeance,
                        permeate_composition=permeate_composition,
                        feed_composition=composition,
                        feed_temperature=feed_temperature,
                        permeate_temperature=permeate_temperature,
                        permeate_pressure=permeate_pressure,
                    )
                )
                d = max(
                    abs(permeate_composition_new.first - permeate_composition.first),
                    abs(permeate_composition_new.second - permeate_composition.second),
                )
            except ():
                raise ValueError(
                    "Partial fluxes are not defined in the stated conditions range"
                )
            else:
                permeate_composition = permeate_composition_new

        return self.get_partial_fluxes_from_permeate_composition(
            first_component_permeance=first_component_permeance,
            second_component_permeance=second_component_permeance,
            permeate_composition=permeate_composition,
            feed_composition=composition,
            feed_temperature=feed_temperature,
            permeate_temperature=permeate_temperature,
            permeate_pressure=permeate_pressure,
        )

    def calculate_permeate_composition(
        self,
        feed_temperature: float,
        composition: Composition,
        precision: typing.Optional[float] = 5e-5,
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
    ) -> Composition:
        """
        Calculates permeate composition at given conditions
        Either permeate temperature or permeate pressure could be stated
        :param feed_temperature: Feed temperature, K
        :param composition: Feed Composition
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :param permeate_temperature: Permeate temperature, if not specified permeate pressure is set to 0 kPa
        :param permeate_pressure - permeate pressure, kPa , if not specified permeate pressure is considered 0 kPa
        :return: Permeate composition in weight %
        """

        x = self.calculate_partial_fluxes(
            feed_temperature,
            composition,
            precision,
            permeate_temperature,
            permeate_pressure,
        )
        return Composition(x[0] / numpy.sum(x), type=CompositionType.weight)

    def calculate_separation_factor(
        self,
        feed_temperature: float,
        composition: Composition,
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
        precision: typing.Optional[float] = 5e-5,
    ) -> float:
        """
        Calculates separation factor at given conditions
        """
        perm_comp = self.calculate_permeate_composition(
            feed_temperature,
            composition,
            precision,
            permeate_temperature,
            permeate_pressure,
        )
        return (composition.second / composition.first) / (
            perm_comp.second / perm_comp.first
        )

    def ideal_diffusion_curve(
        self,
        feed_temperature: float,
        compositions: typing.List[Composition],
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
        precision: typing.Optional[float] = 5e-5,
    ) -> DiffusionCurve:
        """
        Models Ideal Diffusion curve of a specified membrane, at a given temperature, for a given Mixture
        if Ideal experiments for both test_components are available.
        Either permeate temperature or permeate pressure could be stated;
        :param feed_temperature: Feed temperature, K
        :param compositions: List of compositions to model parameters at
        :param permeate_temperature: Permeate temperature, if not specified permeate pressure is set to 0 kPa
        :param permeate_pressure - Permeate pressure, kPa , if not specified permeate pressure is considered 0 kPa
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :return: A DiffusionCurve Object
        """
        return DiffusionCurve(
            mixture=self.mixture,
            membrane_name=self.membrane.name,
            feed_temperature=feed_temperature,
            permeate_temperature=permeate_temperature,
            permeate_pressure=permeate_pressure,
            feed_compositions=compositions,
            partial_fluxes=[
                self.calculate_partial_fluxes(
                    feed_temperature,
                    composition,
                    precision,
                    permeate_temperature,
                    permeate_pressure,
                )
                for composition in compositions
            ],
            comments=(
                str(self.membrane.name)
                + " "
                + str(self.mixture.first_component.name)
                + " / "
                + str(self.mixture.second_component.name)
                + " "
                + datetime.now().strftime("%m/%d/%Y, %H:%M")
            ),
        )

    def ideal_isothermal_process(
        self,
        number_of_steps: int,
        delta_hours: float,
        conditions: Conditions,
        precision: typing.Optional[float] = 5e-5,
    ) -> ProcessModel:
        """
        Models mass and heat balance of an Ideal (constant Permeance) Isothermal Pervaporation Process
        :param number_of_steps: Number of time steps to include in the model
        :param delta_hours: The duration of each step in hours
        :param conditions: Conditions object, where initial conditions are specified
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :return: A ProcessModel Object
        """

        time: typing.List[float] = [
            delta_hours * step for step in range(number_of_steps)
        ]

        partial_fluxes: typing.List[typing.Tuple[float, float]] = []

        first_component_permeance = self.membrane.get_permeance(
            conditions.initial_feed_temperature, self.mixture.first_component
        )
        second_component_permeance = self.membrane.get_permeance(
            conditions.initial_feed_temperature, self.mixture.second_component
        )
        permeances: typing.List[typing.Tuple[Permeance, Permeance]] = [
            (
                first_component_permeance,
                second_component_permeance,
            )
        ] * number_of_steps

        permeate_composition: typing.List[Composition] = []
        feed_composition: typing.List[Composition] = [
            conditions.initial_feed_composition.to_weight(self.mixture)
        ]

        feed_evaporation_heat: typing.List[float] = []
        permeate_condensation_heat: typing.List[typing.Optional[float]] = []
        feed_mass: typing.List[float] = [conditions.initial_feed_amount]

        evaporation_heat_1 = (
            self.mixture.first_component.get_vaporisation_heat(
                conditions.initial_feed_temperature
            )
            / self.mixture.first_component.molecular_weight
            * 1000
        )
        evaporation_heat_2 = (
            self.mixture.second_component.get_vaporisation_heat(
                conditions.initial_feed_temperature
            )
            / self.mixture.first_component.molecular_weight
            * 1000
        )
        if conditions.permeate_temperature is None:
            condensation_heat_1 = None
            condensation_heat_2 = None
            cooling_heat_1 = None
            cooling_heat_2 = None

        else:
            condensation_heat_1 = (
                self.mixture.first_component.get_vaporisation_heat(
                    conditions.permeate_temperature
                )
                / self.mixture.first_component.molecular_weight
                * 1000
            )
            condensation_heat_2 = (
                self.mixture.first_component.get_vaporisation_heat(
                    conditions.permeate_temperature
                )
                / self.mixture.first_component.molecular_weight
                * 1000
            )
            cooling_heat_1 = self.mixture.first_component.get_cooling_heat(
                conditions.permeate_temperature, conditions.initial_feed_temperature
            )
            cooling_heat_2 = self.mixture.second_component.get_cooling_heat(
                conditions.permeate_temperature, conditions.initial_feed_temperature
            )

        for step in range(len(time)):
            partial_fluxes.append(
                self.calculate_partial_fluxes(
                    feed_temperature=conditions.initial_feed_temperature,
                    composition=feed_composition[step],
                    precision=precision,
                    permeate_temperature=conditions.permeate_temperature,
                    permeate_pressure=conditions.permeate_pressure,
                    first_component_permeance=first_component_permeance,
                    second_component_permeance=second_component_permeance,
                )
            )

            permeate_composition.append(
                Composition(
                    p=partial_fluxes[step][0] / (sum(partial_fluxes[step])),
                    type=CompositionType.weight,
                )
            )

            d_mass_1 = partial_fluxes[step][0] * conditions.membrane_area * delta_hours
            d_mass_2 = partial_fluxes[step][1] * conditions.membrane_area * delta_hours

            feed_evaporation_heat.append(
                evaporation_heat_1 * d_mass_1 + evaporation_heat_2 * d_mass_2
            )
            if conditions.permeate_temperature is None:
                permeate_condensation_heat.append(None)
            else:
                permeate_condensation_heat.append(
                    condensation_heat_1 * d_mass_1
                    + condensation_heat_2 * d_mass_2
                    + (cooling_heat_1 * d_mass_1 + cooling_heat_2 * d_mass_2)
                    * (
                        conditions.initial_feed_temperature
                        - conditions.permeate_temperature
                    )
                )

            feed_mass.append(feed_mass[step] - d_mass_1 - d_mass_2)

            feed_composition.append(
                Composition(
                    p=(feed_composition[step].p * feed_mass[step] - d_mass_1)
                    / feed_mass[step + 1],
                    type=CompositionType.weight,
                )
            )

        feed_composition.pop(-1)
        feed_mass.pop(-1)

        return ProcessModel(
            mixture=self.mixture,
            membrane_name=self.membrane.name,
            feed_temperature=[conditions.initial_feed_temperature] * number_of_steps,
            feed_compositions=feed_composition,
            permeate_composition=permeate_composition,
            permeate_temperature=[conditions.permeate_temperature] * number_of_steps,
            permeate_pressure=[conditions.permeate_pressure] * number_of_steps,
            feed_mass=feed_mass,
            partial_fluxes=partial_fluxes,
            permeances=permeances,
            time=time,
            feed_evaporation_heat=feed_evaporation_heat,
            permeate_condensation_heat=permeate_condensation_heat,
            initial_conditions=conditions,
            comments=(
                f"{self.membrane.name} {self.mixture.first_component.name} / {self.mixture.second_component.name}"
                f"Ideal Process Model" + datetime.now().strftime("%m/%d/%Y, %H:%M")
            ),
            membrane_path=self.membrane.path,
        )

    def ideal_non_isothermal_process(
        self,
        conditions: Conditions,
        number_of_steps: int,
        delta_hours: float,
        precision: typing.Optional[float] = 5e-5,
    ) -> ProcessModel:
        """
        Models mass and heat balance of an Ideal (constant Permeance) Non-Isothermal Pervaporation Process.
        The temperature program maybe specified in Conditions, by including a TemperatureProgram object.
        If temperature program is not specified, models self-cooling process;
        :param number_of_steps: Number of time steps to include in the model
        :param delta_hours: The duration of each step in hours
        :param conditions: Conditions object, where initial conditions are specified
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :return: A ProcessModel Object
        """

        time: typing.List[float] = [
            delta_hours * step for step in range(number_of_steps)
        ]

        feed_temperature: typing.List[float] = [conditions.initial_feed_temperature]

        partial_fluxes: typing.List[typing.Tuple[float, float]] = []

        permeances: typing.List[typing.Tuple[Permeance, Permeance]] = []

        permeate_composition: typing.List[Composition] = []

        feed_composition: typing.List[Composition] = [
            conditions.initial_feed_composition.to_weight(self.mixture)
        ]

        feed_evaporation_heat: typing.List[float] = []

        permeate_condensation_heat: typing.List[typing.Optional[float]] = []

        feed_mass: typing.List[float] = [conditions.initial_feed_amount]

        for step in range(len(time)):

            evaporation_heat_1 = (
                self.mixture.first_component.get_vaporisation_heat(
                    feed_temperature[step]
                )
                / self.mixture.first_component.molecular_weight
                * 1000
            )
            evaporation_heat_2 = (
                self.mixture.second_component.get_vaporisation_heat(
                    feed_temperature[step]
                )
                / self.mixture.second_component.molecular_weight
                * 1000
            )

            heat_capacity_1 = (
                self.mixture.first_component.get_specific_heat(feed_temperature[step])
                / self.mixture.first_component.molecular_weight
            )
            heat_capacity_2 = (
                self.mixture.second_component.get_specific_heat(feed_temperature[step])
                / self.mixture.second_component.molecular_weight
            )
            feed_heat_capacity = (
                feed_composition[step].first * heat_capacity_1
                + feed_composition[step].second * heat_capacity_2
            )

            permeances.append(
                (
                    self.membrane.get_permeance(
                        feed_temperature[step], self.mixture.first_component
                    ),
                    self.membrane.get_permeance(
                        feed_temperature[step], self.mixture.second_component
                    ),
                )
            )

            partial_fluxes.append(
                self.calculate_partial_fluxes(
                    feed_temperature=feed_temperature[step],
                    composition=feed_composition[step],
                    precision=precision,
                    permeate_temperature=conditions.permeate_temperature,
                    permeate_pressure=conditions.permeate_pressure,
                    first_component_permeance=permeances[step][0],
                    second_component_permeance=permeances[step][1],
                )
            )

            permeate_composition.append(
                Composition(
                    p=partial_fluxes[step][0] / (sum(partial_fluxes[step])),
                    type=CompositionType.weight,
                )
            )

            d_mass_1 = partial_fluxes[step][0] * conditions.membrane_area * delta_hours
            d_mass_2 = partial_fluxes[step][1] * conditions.membrane_area * delta_hours

            if conditions.permeate_temperature is None:
                permeate_condensation_heat.append(None)
            else:
                condensation_heat_1 = (
                    self.mixture.first_component.get_vaporisation_heat(
                        conditions.permeate_temperature
                    )
                    / self.mixture.first_component.molecular_weight
                    * 1000
                )
                condensation_heat_2 = (
                    self.mixture.second_component.get_vaporisation_heat(
                        conditions.permeate_temperature
                    )
                    / self.mixture.second_component.molecular_weight
                    * 1000
                )

                specific_heat_1 = self.mixture.first_component.get_cooling_heat(
                    feed_temperature[step], conditions.permeate_temperature
                )
                specific_heat_2 = self.mixture.second_component.get_cooling_heat(
                    feed_temperature[step], conditions.permeate_temperature
                )

                permeate_condensation_heat.append(
                    condensation_heat_1 * d_mass_1
                    + condensation_heat_2 * d_mass_2
                    + (specific_heat_1 * d_mass_1 + specific_heat_2 * d_mass_2)
                    * (feed_temperature[step] - conditions.permeate_temperature)
                )

            feed_evaporation_heat.append(
                evaporation_heat_1 * d_mass_1 + evaporation_heat_2 * d_mass_2
            )

            feed_mass.append(feed_mass[step] - d_mass_1 - d_mass_2)

            feed_composition.append(
                Composition(
                    p=(feed_composition[step].p * feed_mass[step] - d_mass_1)
                    / feed_mass[step + 1],
                    type=CompositionType.weight,
                )
            )

            if conditions.temperature_program is None:
                feed_temperature.append(
                    feed_temperature[step]
                    - (
                        feed_evaporation_heat[step]
                        / (feed_heat_capacity * feed_mass[step])
                    )
                )
            else:
                feed_temperature.append(
                    conditions.temperature_program.program(time[step] + delta_hours)
                )

        feed_mass.pop(-1)
        feed_composition.pop(-1)
        feed_temperature.pop(-1)

        return ProcessModel(
            mixture=self.mixture,
            membrane_name=self.membrane.name,
            feed_temperature=feed_temperature,
            permeate_temperature=[conditions.permeate_temperature] * number_of_steps,
            permeate_pressure=[conditions.permeate_pressure] * number_of_steps,
            feed_compositions=feed_composition,
            permeate_composition=permeate_composition,
            feed_mass=feed_mass,
            partial_fluxes=partial_fluxes,
            permeances=permeances,
            time=time,
            feed_evaporation_heat=feed_evaporation_heat,
            permeate_condensation_heat=permeate_condensation_heat,
            initial_conditions=conditions,
            comments=(
                f"{self.membrane.name} {self.mixture.first_component.name} / {self.mixture.second_component.name}"
                f"Ideal Process Model" + datetime.now().strftime("%m/%d/%Y, %H:%M")
            ),
            membrane_path=self.membrane.path,
        )

    def non_ideal_diffusion_curve(
        self,
        diffusion_curve_set: DiffusionCurveSet,
        feed_temperature: float,
        initial_feed_composition: Composition,
        delta_composition: float,
        number_of_steps: int,
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
        initial_permeances: typing.Optional[typing.Tuple[Permeance, Permeance]] = None,
        precision: typing.Optional[float] = 5e-5,
        n_first: typing.Optional[int] = None,
        n_second: typing.Optional[int] = None,
        m_first: typing.Optional[int] = None,
        m_second: typing.Optional[int] = None,
        include_zero: bool = False,
    ):
        """
        The Fucntion models Non-Ideal Diffusion curve
        Based on a set of Diffusion curves measured at different temperatures;
        The modelling could be also performed based on a single diffusion curve:
        In that case the apparent activation energy of transport is considered constant
        and is calculated for each component based on the IdealExperiments default_membranes provided for the Membrane.
        :param diffusion_curve_set: DiffusionCurveSet used as a basis for fitting of the PervaporationFunctions
        :param feed_temperature: temperature of the feed mixture
        :param initial_feed_composition: initial composition of the feed
        :param delta_composition: a composition step size
        :param number_of_steps: number of modelling steps
        :param permeate_temperature: temperature of permeate
        :param permeate_pressure: pressure of permeate
        :param initial_permeances: initial permeances of components, for calculation of an FR parameter using
        data from DiffusionCurveSet object
        :param precision: precision of permeate composition calculation
        :param n_first: n parameter of the PervaporationFunction of the first component
        :param n_second: n parameter of the PervaporationFunction of the second component
        :param m_first: m parameter of the PervaporationFunction of the first component
        :param m_second: m parameter of the PervaporationFunction of the second component
        :param include_zero: bool parameter to force default points while fitting the PervaporationFunction
        :return: non-ideal diffusion curve
        """

        measurements_first = Measurements.from_diffusion_curves_first(
            diffusion_curve_set
        )
        measurements_second = Measurements.from_diffusion_curves_second(
            diffusion_curve_set
        )

        if len(diffusion_curve_set.diffusion_curves) == 1:
            pervaporation_function_temperature = diffusion_curve_set.diffusion_curves[
                0
            ].feed_temperature

            pervaporation_function_first = find_best_fit(
                data=measurements_first,
                n=n_first,
                m=0,
                include_zero=include_zero,
                component_index=0,
            )

            pervaporation_function_second = find_best_fit(
                data=measurements_second,
                n=n_second,
                m=0,
                include_zero=include_zero,
                component_index=1,
            )

            if pervaporation_function_temperature == feed_temperature:
                pass
            else:
                activation_energy_first = self.membrane.calculate_activation_energy(
                    self.mixture.first_component
                )
                activation_energy_second = self.membrane.calculate_activation_energy(
                    self.mixture.second_component
                )

                pervaporation_function_first = pervaporation_function_first * numpy.exp(
                    -pervaporation_function_first.b[0]
                    / pervaporation_function_temperature
                    + activation_energy_first / (R * pervaporation_function_temperature)
                )

                pervaporation_function_second = (
                    pervaporation_function_second
                    * numpy.exp(
                        -pervaporation_function_second.b[0]
                        / pervaporation_function_temperature
                        + activation_energy_second
                        / (R * pervaporation_function_temperature)
                    )
                )

                pervaporation_function_first.b[0] = activation_energy_first / R
                pervaporation_function_second.b[0] = activation_energy_second / R

        else:
            pervaporation_function_first = find_best_fit(
                data=measurements_first,
                n=n_first,
                m=m_first,
                include_zero=include_zero,
                component_index=0,
            )
            pervaporation_function_second = find_best_fit(
                data=measurements_second,
                n=n_second,
                m=m_second,
                include_zero=include_zero,
                component_index=1,
            )

        if initial_permeances is None:
            first_component_permeance = Permeance(
                value=pervaporation_function_first(
                    initial_feed_composition.first, feed_temperature
                )
            )

            second_component_permeance = Permeance(
                value=pervaporation_function_second(
                    initial_feed_composition.first, feed_temperature
                )
            )

        else:
            first_component_permeance = initial_permeances[0].convert(
                to_units=Units.kg_m2_h_kPa, component=self.mixture.first_component
            )
            second_component_permeance = initial_permeances[1].convert(
                to_units=Units.kg_m2_h_kPa, component=self.mixture.second_component
            )

        permeances: typing.List[typing.Tuple[Permeance, Permeance]] = [
            (first_component_permeance, second_component_permeance)
        ]
        facilitation_rate_first = (
            first_component_permeance.value
            / pervaporation_function_first(
                x=initial_feed_composition.first, t=feed_temperature
            )
        )
        facilitation_rate_second = (
            second_component_permeance.value
            / pervaporation_function_second(
                x=initial_feed_composition.first, t=feed_temperature
            )
        )

        partial_fluxes: typing.List[typing.Tuple[float, float]] = []
        compositions: typing.List[Composition] = [
            initial_feed_composition.to_weight(self.mixture)
        ]

        for i in range(number_of_steps + 1):

            compositions.append(
                Composition(
                    p=compositions[i].first + delta_composition,
                    type=CompositionType.weight,
                )
            )

            partial_fluxes.append(
                self.calculate_partial_fluxes(
                    feed_temperature=feed_temperature,
                    composition=compositions[i],
                    precision=precision,
                    permeate_temperature=permeate_temperature,
                    permeate_pressure=permeate_pressure,
                    first_component_permeance=permeances[i][0],
                    second_component_permeance=permeances[i][1],
                )
            )

            permeances.append(
                (
                    Permeance(
                        value=(
                            pervaporation_function_first(
                                compositions[i + 1].first, feed_temperature
                            )
                            * facilitation_rate_first
                        )
                    ),
                    Permeance(
                        value=(
                            pervaporation_function_second(
                                compositions[i + 1].first, feed_temperature
                            )
                            * facilitation_rate_second
                        )
                    ),
                )
            )

        compositions.pop(-1)
        permeances.pop(-1)

        return DiffusionCurve(
            mixture=self.mixture,
            membrane_name=self.membrane.name,
            feed_temperature=feed_temperature,
            permeate_temperature=permeate_temperature,
            permeate_pressure=permeate_pressure,
            feed_compositions=compositions,
            partial_fluxes=partial_fluxes,
            permeances=permeances,
            comments=(
                f"{self.membrane.name} {self.mixture.first_component.name} "
                f"/ {self.mixture.second_component.name}"
                + datetime.now().strftime("%m/%d/%Y, %H:%M")
            ),
        )

    def non_ideal_isothermal_process(
        self,
        conditions: Conditions,
        diffusion_curve_set: DiffusionCurveSet,
        number_of_steps: int,
        delta_hours: float,
        precision: typing.Optional[float] = 5e-5,
        initial_permeances: typing.Optional[typing.Tuple[Permeance, Permeance]] = None,
        n_first: typing.Optional[int] = None,
        m_first: typing.Optional[int] = None,
        n_second: typing.Optional[int] = None,
        m_second: typing.Optional[int] = None,
        include_zero: bool = False,
    ):
        """
        The function models Non-Ideal Isothermal Process
        Based on a set of Diffusion curves measured at different temperatures;
        The modelling could be also performed based on a single diffusion curve:
        In that case the apparent activation energy of transport is considered constant
        and is calculated for each component based on the IdealExperiments default_membranes provided for the Membrane.
        :param conditions: Initial Conditions of the Process
        :param diffusion_curve_set: A set of Diffusion curves picked for the Modelling form the Membrane
        :param number_of_steps: Number of time steps for modelling
        :param delta_hours: Size of each step in hours
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :param initial_permeances: Initial Permeances, should be stated if the Membranes swelling history is significant
        :param n_first: Optional parameter,
        indicates the order of the polynomial of the composition part of the Permeance function for the
        first component
        :param m_first: Optional parameter,
        indicates the order of the polynomial of the temperature part of the Permeance function for
        the first component
        :param n_second: Optional parameter,
        indicates the order of the polynomial of the composition part of the Permeance function for
        the second component
        :param m_second: Optional parameter,
        indicates the order of the polynomial of the temperature part of the Permeance function for
        the second component
        :param include_zero: if True, points:
         first_component_fraction = 0 first_component_permeance=0 for the first test_components
         first_component_fraction = 1 second_component_permeance=0 for the second test_components
         for each temperature are added to the measurements in order to improve obtained fits
        :return: ProcessModel object
        """
        for curve in diffusion_curve_set.diffusion_curves:
            for c in curve.feed_compositions:
                if c.type == CompositionType.molar:
                    c.to_weight(curve.mixture)

        time: typing.List[float] = [
            delta_hours * step for step in range(number_of_steps)
        ]

        partial_fluxes: typing.List[typing.Tuple[float, float]] = []

        permeate_composition: typing.List[Composition] = []
        feed_composition: typing.List[Composition] = [
            conditions.initial_feed_composition.to_weight(self.mixture)
        ]

        feed_evaporation_heat: typing.List[float] = []
        permeate_condensation_heat: typing.List[typing.Optional[float]] = []
        feed_mass: typing.List[float] = [conditions.initial_feed_amount]

        measurements_first = Measurements.from_diffusion_curves_first(
            diffusion_curve_set
        )
        measurements_second = Measurements.from_diffusion_curves_second(
            diffusion_curve_set
        )

        if len(diffusion_curve_set.diffusion_curves) == 1:
            pervaporation_function_temperature = diffusion_curve_set.diffusion_curves[
                0
            ].feed_temperature

            pervaporation_function_first = find_best_fit(
                data=measurements_first,
                n=n_first,
                m=0,
                include_zero=False,
                component_index=0,
            )
            pervaporation_function_second = find_best_fit(
                data=measurements_second,
                n=n_second,
                m=0,
                include_zero=False,
                component_index=1,
            )

            if (
                pervaporation_function_temperature
                == conditions.initial_feed_temperature
            ):
                pass
            else:
                activation_energy_first = self.membrane.calculate_activation_energy(
                    self.mixture.first_component
                )
                activation_energy_second = self.membrane.calculate_activation_energy(
                    self.mixture.second_component
                )

                pervaporation_function_first = pervaporation_function_first * numpy.exp(
                    -pervaporation_function_first.b[0]
                    / pervaporation_function_temperature
                    + activation_energy_first / (R * pervaporation_function_temperature)
                )

                pervaporation_function_second = (
                    pervaporation_function_second
                    * numpy.exp(
                        -pervaporation_function_second.b[0]
                        / pervaporation_function_temperature
                        + activation_energy_second
                        / (R * pervaporation_function_temperature)
                    )
                )

                pervaporation_function_first.b[0] = activation_energy_first / R
                pervaporation_function_second.b[0] = activation_energy_second / R

        else:
            pervaporation_function_first = find_best_fit(
                data=measurements_first,
                n=n_first,
                m=m_first,
                include_zero=include_zero,
                component_index=0,
            )
            pervaporation_function_second = find_best_fit(
                data=measurements_second,
                n=n_second,
                m=m_second,
                include_zero=include_zero,
                component_index=1,
            )

        if initial_permeances is None:
            first_component_permeance = Permeance(
                value=pervaporation_function_first(
                    feed_composition[0].first, conditions.initial_feed_temperature
                )
            )

            second_component_permeance = Permeance(
                value=pervaporation_function_second(
                    feed_composition[0].first, conditions.initial_feed_temperature
                )
            )

        else:
            first_component_permeance = initial_permeances[0].convert(
                to_units=Units.kg_m2_h_kPa, component=self.mixture.first_component
            )
            second_component_permeance = initial_permeances[1].convert(
                to_units=Units.kg_m2_h_kPa, component=self.mixture.second_component
            )

        permeances: typing.List[typing.Tuple[Permeance, Permeance]] = [
            (first_component_permeance, second_component_permeance)
        ]

        facilitation_rate_first = (
            first_component_permeance.value
            / pervaporation_function_first(
                x=conditions.initial_feed_composition.first,
                t=conditions.initial_feed_temperature,
            )
        )
        facilitation_rate_second = (
            second_component_permeance.value
            / pervaporation_function_second(
                x=conditions.initial_feed_composition.first,
                t=conditions.initial_feed_temperature,
            )
        )

        evaporation_heat_1 = (
            self.mixture.first_component.get_vaporisation_heat(
                conditions.initial_feed_temperature
            )
            / self.mixture.first_component.molecular_weight
            * 1000
        )
        evaporation_heat_2 = (
            self.mixture.second_component.get_vaporisation_heat(
                conditions.initial_feed_temperature
            )
            / self.mixture.first_component.molecular_weight
            * 1000
        )
        if conditions.permeate_temperature is None:
            condensation_heat_1 = None
            condensation_heat_2 = None
            cooling_heat_1 = None
            cooling_heat_2 = None

        else:
            condensation_heat_1 = (
                self.mixture.first_component.get_vaporisation_heat(
                    conditions.permeate_temperature
                )
                / self.mixture.first_component.molecular_weight
                * 1000
            )
            condensation_heat_2 = (
                self.mixture.first_component.get_vaporisation_heat(
                    conditions.permeate_temperature
                )
                / self.mixture.first_component.molecular_weight
                * 1000
            )
            cooling_heat_1 = self.mixture.first_component.get_cooling_heat(
                conditions.permeate_temperature, conditions.initial_feed_temperature
            )
            cooling_heat_2 = self.mixture.second_component.get_cooling_heat(
                conditions.permeate_temperature, conditions.initial_feed_temperature
            )

        for step in range(len(time)):

            partial_fluxes.append(
                self.calculate_partial_fluxes(
                    feed_temperature=conditions.initial_feed_temperature,
                    composition=feed_composition[step],
                    precision=precision,
                    permeate_temperature=conditions.permeate_temperature,
                    permeate_pressure=conditions.permeate_pressure,
                    first_component_permeance=permeances[step][0],
                    second_component_permeance=permeances[step][1],
                )
            )

            permeate_composition.append(
                Composition(
                    p=partial_fluxes[step][0] / (sum(partial_fluxes[step])),
                    type=CompositionType.weight,
                )
            )

            d_mass_1 = partial_fluxes[step][0] * conditions.membrane_area * delta_hours
            d_mass_2 = partial_fluxes[step][1] * conditions.membrane_area * delta_hours

            feed_evaporation_heat.append(
                evaporation_heat_1 * d_mass_1 + evaporation_heat_2 * d_mass_2
            )
            if conditions.permeate_temperature is None:
                permeate_condensation_heat.append(None)
            else:
                permeate_condensation_heat.append(
                    condensation_heat_1 * d_mass_1
                    + condensation_heat_2 * d_mass_2
                    + (cooling_heat_1 * d_mass_1 + cooling_heat_2 * d_mass_2)
                    * (
                        conditions.initial_feed_temperature
                        - conditions.permeate_temperature
                    )
                )

            feed_mass.append(feed_mass[step] - d_mass_1 - d_mass_2)

            feed_composition.append(
                Composition(
                    p=(feed_composition[step].p * feed_mass[step] - d_mass_1)
                    / feed_mass[step + 1],
                    type=CompositionType.weight,
                )
            )
            permeances.append(
                (
                    Permeance(
                        value=pervaporation_function_first(
                            x=feed_composition[step].first,
                            t=conditions.initial_feed_temperature,
                        )
                        * facilitation_rate_first
                    ),
                    Permeance(
                        value=pervaporation_function_second(
                            x=feed_composition[step].first,
                            t=conditions.initial_feed_temperature,
                        )
                        * facilitation_rate_second
                    ),
                ),
            )

        feed_composition.pop(-1)
        feed_mass.pop(-1)
        permeances.pop(-1)

        return ProcessModel(
            mixture=self.mixture,
            membrane_name=self.membrane.name,
            feed_temperature=[conditions.initial_feed_temperature] * number_of_steps,
            feed_compositions=feed_composition,
            permeate_composition=permeate_composition,
            permeate_temperature=[conditions.permeate_temperature] * number_of_steps,
            permeate_pressure=[conditions.permeate_pressure] * number_of_steps,
            feed_mass=feed_mass,
            partial_fluxes=partial_fluxes,
            permeances=permeances,
            time=time,
            feed_evaporation_heat=feed_evaporation_heat,
            permeate_condensation_heat=permeate_condensation_heat,
            initial_conditions=conditions,
            permeance_fits=(
                pervaporation_function_first,
                pervaporation_function_second,
            ),
            comments=(
                f"{self.membrane.name} {self.mixture.first_component.name} / {self.mixture.second_component.name}"
                f"Non-ideal Process Model" + datetime.now().strftime("%m/%d/%Y, %H:%M")
            ),
        )

    def non_ideal_non_isothermal_process(
        self,
        conditions: Conditions,
        diffusion_curve_set: DiffusionCurveSet,
        number_of_steps: int,
        delta_hours: float,
        precision: typing.Optional[float] = 5e-5,
        initial_permeances: typing.Optional[typing.Tuple[Permeance, Permeance]] = None,
        n_first: typing.Optional[int] = None,
        m_first: typing.Optional[int] = None,
        n_second: typing.Optional[int] = None,
        m_second: typing.Optional[int] = None,
        include_zero: bool = False,
    ) -> ProcessModel:
        """
        The function models Non-Ideal Non-Isothermal Process
        Based on a set of Diffusion curves measured at different temperatures;
        The modelling could be also performed based on a single diffusion curve:
        In that case the apparent activation energy of transport is considered constant
        and is calculated for each test_components based on the IdealExperiments default_membranes provided for the
        Membrane.
        :param conditions: Initial Conditions of the Process, a Temperature program may be added if necessary
        :param diffusion_curve_set: A set of Diffusion curves picked for the Modelling from the Membrane
        :param number_of_steps: Number of time steps for modelling
        :param delta_hours: Size of each step in hours
        :param precision: Precision in obtained permeate composition, by default is 5e-5
        :param initial_permeances: Initial Permeances, should be stated if the Membranes swelling history is significant
        :param n_first: Optional parameter,
        indicates the order of the polynomial of the composition part of the Permeance function for
         the first component
        :param m_first: Optional parameter,
        indicates the order of the polynomial of the temperature part of the Permeance function for
         the first component
        :param n_second: Optional parameter,
        indicates the order of the polynomial of the composition part of the Permeance function for
         the second component
        :param m_second: Optional parameter,
        indicates the order of the polynomial of the temperature part of the Permeance function for
         the second component
        :param include_zero: if True, points:
         first_component_fraction = 0 first_component_permeance=0 for the first test_components
         first_component_fraction = 1 second_component_permeance=0 for the second test_components
         for each temperature are added to the measurements in order to improve obtained fits
        :return: ProcessModel object
        """
        for curve in diffusion_curve_set.diffusion_curves:
            for c in curve.feed_compositions:
                if c.type == CompositionType.molar:
                    c.to_weight(curve.mixture)

        time: typing.List[float] = [
            delta_hours * step for step in range(number_of_steps)
        ]

        feed_temperature: typing.List[float] = [conditions.initial_feed_temperature]

        partial_fluxes: typing.List[typing.Tuple[float, float]] = []

        permeate_composition: typing.List[Composition] = []

        feed_composition: typing.List[Composition] = [
            conditions.initial_feed_composition.to_weight(self.mixture)
        ]

        feed_evaporation_heat: typing.List[float] = []

        permeate_condensation_heat: typing.List[typing.Optional[float]] = []

        feed_mass: typing.List[float] = [conditions.initial_feed_amount]

        measurements_first = Measurements.from_diffusion_curves_first(
            diffusion_curve_set
        )
        measurements_second = Measurements.from_diffusion_curves_second(
            diffusion_curve_set
        )

        if len(diffusion_curve_set.diffusion_curves) == 1:
            pervaporation_function_temperature = diffusion_curve_set.diffusion_curves[
                0
            ].feed_temperature
            activation_energy_first = self.membrane.calculate_activation_energy(
                self.mixture.first_component
            )
            activation_energy_second = self.membrane.calculate_activation_energy(
                self.mixture.second_component
            )

            pervaporation_function_first = find_best_fit(
                data=measurements_first,
                n=n_first,
                m=0,
                include_zero=False,
                component_index=0,
            )
            pervaporation_function_second = find_best_fit(
                data=measurements_second,
                n=n_second,
                m=0,
                include_zero=False,
                component_index=1,
            )

            pervaporation_function_first = pervaporation_function_first * numpy.exp(
                -pervaporation_function_first.b[0] / pervaporation_function_temperature
                + activation_energy_first / (R * pervaporation_function_temperature)
            )

            pervaporation_function_second = pervaporation_function_second * numpy.exp(
                -pervaporation_function_second.b[0] / pervaporation_function_temperature
                + activation_energy_second / (R * pervaporation_function_temperature)
            )

            pervaporation_function_first.b[0] = activation_energy_first / R
            pervaporation_function_second.b[0] = activation_energy_second / R

        else:
            pervaporation_function_first = find_best_fit(
                data=measurements_first,
                n=n_first,
                m=m_first,
                include_zero=include_zero,
                component_index=0,
            )
            pervaporation_function_second = find_best_fit(
                data=measurements_second,
                n=n_second,
                m=m_second,
                include_zero=include_zero,
                component_index=1,
            )

        if initial_permeances is None:
            first_component_permeance = Permeance(
                value=pervaporation_function_first(
                    feed_composition[0].p, conditions.initial_feed_temperature
                )
            )

            second_component_permeance = Permeance(
                value=pervaporation_function_second(
                    feed_composition[0].p, conditions.initial_feed_temperature
                )
            )

        else:
            first_component_permeance = initial_permeances[0].convert(
                to_units=Units.kg_m2_h_kPa, component=self.mixture.first_component
            )
            second_component_permeance = initial_permeances[1].convert(
                to_units=Units.kg_m2_h_kPa, component=self.mixture.second_component
            )

        permeances: typing.List[typing.Tuple[Permeance, Permeance]] = [
            (first_component_permeance, second_component_permeance)
        ]

        facilitation_rate_first = (
            first_component_permeance.value
            / pervaporation_function_first(
                x=conditions.initial_feed_composition.first,
                t=conditions.initial_feed_temperature,
            )
        )
        facilitation_rate_second = (
            second_component_permeance.value
            / pervaporation_function_second(
                x=conditions.initial_feed_composition.first,
                t=conditions.initial_feed_temperature,
            )
        )

        for step in range(len(time)):

            evaporation_heat_1 = (
                self.mixture.first_component.get_vaporisation_heat(
                    feed_temperature[step]
                )
                / self.mixture.first_component.molecular_weight
                * 1000
            )
            evaporation_heat_2 = (
                self.mixture.second_component.get_vaporisation_heat(
                    feed_temperature[step]
                )
                / self.mixture.second_component.molecular_weight
                * 1000
            )

            heat_capacity_1 = (
                self.mixture.first_component.get_specific_heat(feed_temperature[step])
                / self.mixture.first_component.molecular_weight
            )
            heat_capacity_2 = (
                self.mixture.second_component.get_specific_heat(feed_temperature[step])
                / self.mixture.second_component.molecular_weight
            )
            feed_heat_capacity = (
                feed_composition[step].first * heat_capacity_1
                + feed_composition[step].second * heat_capacity_2
            )

            partial_fluxes.append(
                self.calculate_partial_fluxes(
                    feed_temperature=feed_temperature[step],
                    composition=feed_composition[step],
                    precision=precision,
                    permeate_temperature=conditions.permeate_temperature,
                    permeate_pressure=conditions.permeate_pressure,
                    first_component_permeance=permeances[step][0],
                    second_component_permeance=permeances[step][1],
                )
            )

            permeate_composition.append(
                Composition(
                    p=partial_fluxes[step][0] / (sum(partial_fluxes[step])),
                    type=CompositionType.weight,
                )
            )

            d_mass_1 = partial_fluxes[step][0] * conditions.membrane_area * delta_hours
            d_mass_2 = partial_fluxes[step][1] * conditions.membrane_area * delta_hours

            if conditions.permeate_temperature is None:
                permeate_condensation_heat.append(None)
            else:
                condensation_heat_1 = (
                    self.mixture.first_component.get_vaporisation_heat(
                        conditions.permeate_temperature
                    )
                    / self.mixture.first_component.molecular_weight
                    * 1000
                )
                condensation_heat_2 = (
                    self.mixture.second_component.get_vaporisation_heat(
                        conditions.permeate_temperature
                    )
                    / self.mixture.second_component.molecular_weight
                    * 1000
                )

                specific_heat_1 = self.mixture.first_component.get_cooling_heat(
                    feed_temperature[step], conditions.permeate_temperature
                )
                specific_heat_2 = self.mixture.second_component.get_cooling_heat(
                    feed_temperature[step], conditions.permeate_temperature
                )

                permeate_condensation_heat.append(
                    condensation_heat_1 * d_mass_1
                    + condensation_heat_2 * d_mass_2
                    + (specific_heat_1 * d_mass_1 + specific_heat_2 * d_mass_2)
                    * (feed_temperature[step] - conditions.permeate_temperature)
                )

            feed_evaporation_heat.append(
                evaporation_heat_1 * d_mass_1 + evaporation_heat_2 * d_mass_2
            )

            feed_mass.append(feed_mass[step] - d_mass_1 - d_mass_2)

            feed_composition.append(
                Composition(
                    p=(feed_composition[step].p * feed_mass[step] - d_mass_1)
                    / feed_mass[step + 1],
                    type=CompositionType.weight,
                )
            )

            if conditions.temperature_program is None:
                feed_temperature.append(
                    feed_temperature[step]
                    - (
                        feed_evaporation_heat[step]
                        / (feed_heat_capacity * feed_mass[step])
                    )
                )
            else:
                feed_temperature.append(
                    conditions.temperature_program.program(time[step] + delta_hours)
                )

            permeances.append(
                (
                    Permeance(
                        value=pervaporation_function_first(
                            x=feed_composition[step + 1].first,
                            t=feed_temperature[step + 1],
                        )
                        * facilitation_rate_first
                    ),
                    Permeance(
                        value=pervaporation_function_second(
                            x=feed_composition[step + 1].first,
                            t=feed_temperature[step + 1],
                        )
                        * facilitation_rate_second
                    ),
                ),
            )

        feed_composition.pop(-1)
        feed_mass.pop(-1)
        permeances.pop(-1)
        feed_temperature.pop(-1)

        return ProcessModel(
            mixture=self.mixture,
            membrane_name=self.membrane.name,
            feed_temperature=feed_temperature,
            permeate_temperature=[conditions.permeate_temperature] * number_of_steps,
            permeate_pressure=[conditions.permeate_pressure] * number_of_steps,
            feed_compositions=feed_composition,
            permeate_composition=permeate_composition,
            feed_mass=feed_mass,
            partial_fluxes=partial_fluxes,
            permeances=permeances,
            time=time,
            feed_evaporation_heat=feed_evaporation_heat,
            permeate_condensation_heat=permeate_condensation_heat,
            initial_conditions=conditions,
            permeance_fits=(
                pervaporation_function_first,
                pervaporation_function_second,
            ),
            comments=(
                f"{self.membrane.name} {self.mixture.first_component.name} / {self.mixture.second_component.name}"
                f"Non-ideal Process Model" + datetime.now().strftime("%m/%d/%Y, %H:%M")
            ),
            membrane_path=self.membrane.path,
        )



import typing

import attr
import numpy

from ..mixtures import Composition


# TODO: Enum or not Enum?
class CalculationType:
    polynomial: str = "polynomial"
    exponential: str = "exponential"
    logarithmic: str = "logarithmic"


@attr.s(auto_attribs=True)
class TemperatureProgram:
    coefficients: typing.List[float]
    type: str = CalculationType.polynomial

    def polynomial(self, x: float) -> float:
        """
        Calculates temperature using polynomial approximation
        :param x: parameter used for calculation of temperature
        :return: Temperature value calculated using polynomial relation defined with .coefficients
        """
        return sum(
            [self.coefficients[i] * x**i for i in range(len(self.coefficients))]
        )

    def exponential(self, x: float) -> float:
        """
        Calculates temperature using exponential approximation
        :param x: parameter used for calculation of temperature
        :return: Temperature value calculated using exponential-polynomial relation defined with .coefficients
        """
        return self.coefficients[0] * numpy.exp(
            sum(
                [
                    self.coefficients[i] * x ** (i - 1)
                    for i in range(1, len(self.coefficients))
                ]
            )
        )

    def logarithmic(self, x: float) -> float:
        """
        Calculates temperature using logarithmic approximation
        :param x: parameter used for calculation of temperature
        :return: Temperature value calculated using logarithmic-polynomial relation defined with .coefficients
        """
        return self.coefficients[0] * numpy.log(
            sum(
                [
                    self.coefficients[i] * x ** (i - 1)
                    for i in range(1, len(self.coefficients))
                ]
            )
        )

    def program(self, time):
        """
        Calculation of the Temperature based on a given temperature program
        :param time - time in hours
        :return - Temperature in K
        """
        return getattr(self, self.type)(time)


@attr.s(auto_attribs=True)
class Conditions:
    """
    A class for specification initial conditions for modelling of the test_pervaporation processes
    """

    membrane_area: float
    initial_feed_temperature: float
    initial_feed_amount: float
    initial_feed_composition: Composition
    permeate_temperature: typing.Optional[float] = None
    permeate_pressure: typing.Optional[float] = None
    temperature_program: typing.Optional[TemperatureProgram] = None



import typing
from pathlib import Path

import attr
import numpy

from ..components import Component
from ..diffusion_curve import DiffusionCurveSet
from ..experiments import IdealExperiments
from ..permeance import Permeance, Units
from ..utils import R


@attr.s(auto_attribs=True)
class Membrane:
    """
    Class to represent a membrane
    """

    name: str
    ideal_experiments: typing.Optional[IdealExperiments] = None
    diffusion_curve_sets: typing.Optional[typing.List[DiffusionCurveSet]] = None

    path: typing.Optional[Path] = None

    @classmethod
    def load(cls, path: typing.Union[str, Path]) -> "Membrane":
        """
        Creates a Membrane object from a directory of a unified type
        :param path: path to the directory, containing infromation on IdealExperiments and DiffusionCutveSets
        :return: a Membrane object
        """
        if type(path) is not Path:
            path = Path(path)

        if (path / "ideal_experiments.csv").exists():
            ie = IdealExperiments.from_csv(path / "ideal_experiments.csv")
        else:
            ie = None

        if (path / "diffusion_curve_sets").exists():
            if len(list((path / "diffusion_curve_sets").iterdir())) > 0:
                diffusion_curve_sets = []
                for file in list((path / "diffusion_curve_sets").iterdir()):
                    if file.stem == ".DS_store":
                        continue
                    diffusion_curve_sets.append(DiffusionCurveSet.load(file))
                if len(diffusion_curve_sets) == 0:
                    diffusion_curve_sets = None
            else:
                diffusion_curve_sets = None
        else:
            diffusion_curve_sets = None

        if ie is None and diffusion_curve_sets is None:
            raise FileExistsError("No default_membranes found at %s" % path)

        results_path = path / "results"
        results_path.mkdir(parents=True, exist_ok=True)

        return cls(
            name=path.stem,
            ideal_experiments=ie,
            diffusion_curve_sets=diffusion_curve_sets,
            path=path,
        )

    def get_penetrant_data(self, component: Component) -> IdealExperiments:
        """
        Gets all ideal experiments of the membrane for a specified Component
        """
        return IdealExperiments(
            experiments=list(
                filter(
                    lambda x: x.component.name == component.name,
                    self.ideal_experiments.experiments,
                )
            )
        )

    def calculate_activation_energy(
        self,
        component: Component,
    ) -> float:
        """
        Calculates Apparent Activation Energy of Transport i J/mol
        for a specified test_components, based on its Ideal Experiments.
        Activation energy is assumed to be independent of concentration for Ideal Process

        """
        component_experiments = self.get_penetrant_data(component)
        stated_activation_energy = component_experiments.experiments[
            0
        ].activation_energy

        if len(component_experiments) < 2 and stated_activation_energy is None:
            raise ValueError(
                "At least two measurements at different temperatures are required for "
                "the calculation of Apparent Activation Energy"
            )
        elif len(component_experiments) < 2 and stated_activation_energy is not None:
            return stated_activation_energy

        x = numpy.divide(
            1,
            [
                ideal_experiment.temperature
                for ideal_experiment in component_experiments.experiments
            ],
        )

        y = numpy.log(
            [
                ideal_experiment.permeance.value
                for ideal_experiment in component_experiments.experiments
            ]
        )

        a = numpy.vstack([x, numpy.ones(len(x))]).T

        activation_energy, c = numpy.linalg.lstsq(a, y, rcond=-1)[0] * R

        return -activation_energy

    def get_permeance(
        self,
        temperature: float,
        component: Component,
        initial_permeance: typing.Optional[Permeance] = None,
    ) -> Permeance:
        """
        Calculates Permeance of a specified test_components (kg/(m2*h*kPa)) at a specified Temperature (K)
        based on a given or calculated Apparent Activation Energy of Transport
        """

        component_experiments = self.get_penetrant_data(component)

        temperature_list = [
            ideal_experiment.temperature
            for ideal_experiment in component_experiments.experiments
        ]

        index = min(
            range(len(temperature_list)),
            key=lambda i: abs(temperature_list[i] - temperature),
        )

        given_permeance = component_experiments.experiments[index].permeance.convert(
            to_units=Units().kg_m2_h_kPa, component=component
        )

        if (
            component_experiments.experiments[index].activation_energy is None
            and component_experiments.experiments[index].temperature != temperature
        ):

            activation_energy = self.calculate_activation_energy(component)
            return Permeance(
                value=given_permeance.value
                * numpy.exp(
                    -activation_energy
                    / R
                    * (1 / temperature - 1 / temperature_list[index])
                )
            )
        elif component_experiments.experiments[index].temperature == temperature:
            return given_permeance

        elif initial_permeance is not None:
            return Permeance(
                value=initial_permeance.value
                * numpy.exp(
                    -component_experiments.experiments[index].activation_energy
                    / R
                    * (1 / temperature - 1 / temperature_list[index])
                )
            )

        else:
            return Permeance(
                value=given_permeance.value
                * numpy.exp(
                    -component_experiments.experiments[index].activation_energy
                    / R
                    * (1 / temperature - 1 / temperature_list[index])
                )
            )

    def get_ideal_selectivity(
        self,
        temperature: float,
        first_component: Component,
        second_component: Component,
        calculation_type: typing.Optional[str] = "molar",
    ) -> float:
        """
        Calculates Ideal selectivity of one specified test_components over the other at a specified temperature.
        """
        if calculation_type == "weight":
            return (
                self.get_permeance(temperature, first_component).value
                / self.get_permeance(temperature, second_component).value
            )
        elif calculation_type == "molar":
            return (
                self.get_permeance(temperature, first_component)
                .convert(to_units=Units().SI, component=first_component)
                .value
                / self.get_permeance(temperature, second_component)
                .convert(to_units=Units().SI, component=second_component)
                .value
            )

    def get_estimated_pure_component_flux(
        self,
        temperature: float,
        component: Component,
        permeate_temperature: typing.Optional[float] = None,
        permeate_pressure: typing.Optional[float] = None,
    ) -> float:
        """
        Calculates Flux of a specified test_components during individual Pervaporation based on the specified Permeance
        from Ideal Experiments Of the Component.
        May be performed either on the basis of permeate temperature in K or permeate pressure in kPa.
        If permeate conditions are not stated partial pressure of the test_components in permeate is considered zero.
        """
        if permeate_temperature is None and permeate_pressure is None:
            return self.get_permeance(
                temperature, component
            ).value * component.get_vapor_pressure(temperature)
        elif permeate_temperature is not None and permeate_pressure is None:
            return self.get_permeance(temperature, component).value * (
                component.get_vapor_pressure(temperature)
                - component.get_vapor_pressure(permeate_temperature)
            )
        elif permeate_pressure is not None and permeate_temperature is None:
            return self.get_permeance(temperature, component).value * (
                component.get_vapor_pressure(temperature) - permeate_pressure
            )
        else:
            raise ValueError(
                "Either permeate temperature or permeate pressure could be stated not both"
            )



from .utils import (HeatCapacityConstants, NRTLParameters, R,
                    VaporPressureConstants, VPConstantsType)

__all__ = [
    "VaporPressureConstants",
    "R",
    "NRTLParameters",
    "HeatCapacityConstants",
    "VPConstantsType",
]



import typing

import attr

from ..components import Component


class Units:
    """
    Class to represent Permeance units
    """

    GPU: str = "GPU"
    SI: str = "SI"
    kg_m2_h_kPa: str = "kg/(m2*h*kPa)"


@attr.s(auto_attribs=True)
class Permeance:
    """
    Class to represent Permeance values
    """

    value: float = attr.ib(converter=lambda x: x if x >= 0 else 0)
    units: str = Units.kg_m2_h_kPa

    def __add__(self, other: "Permeance") -> "Permeance":
        if self.units != other.units:
            raise ValueError("Only Permeances in same units could be added")
        return Permeance(value=self.value + other.value, units=self.units)

    def convert(
        self,
        to_units: str,
        component: typing.Optional[Component] = None,
    ) -> "Permeance":
        """
        Converts Permeance values from and to commonly used units:
        kg/(m2*h*kPa),
        SI - (mol/(m2*s*Pa))
        GPU
        """
        if to_units == self.units:
            return self

        if component is None and to_units == Units().kg_m2_h_kPa:
            raise ValueError(
                "For conversion from and to kg/(m2*h*kPa) Component must be specified "
            )
        try:
            if component is None:
                conversion_dict = {
                    "GPU": 3.35e-10,
                    "SI": 1,
                }
            else:
                conversion_dict = {
                    "GPU": 3.35e-10,
                    "SI": 1,
                    "kg/(m2*h*kPa)": 1 / (component.molecular_weight * 3.6e3),
                }
        except KeyError:
            raise KeyError("Conversion to specified units is not supported")

        return Permeance(
            value=(
                self.value * conversion_dict[self.units] / conversion_dict[to_units]
            ),
            units=to_units,
        )
